from learning_rate_getters import LearningRateGetterConstant
from mappers import MapperLinear, MapperMultiLayer, MapperNeuralNet
from optimizers import OptimizerGradientDescentHeavyBall, OptimizerGradientDescentLineSearch, OptimizerStochasticGradient
from fun_obj import FunObjLeastSquaresMultiOutput
from generic_models import GenericSupervisedModel
from numpy.testing._private.utils import assert_almost_equal, assert_array_almost_equal
from nampy import Constant, Variable
import nampy
import os
import pickle
import argparse
import numpy as np



def load_dataset(filename):
    with open(os.path.join('..','data',filename), 'rb') as f:
        return pickle.load(f)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-q','--question', required=True)

    io_args = parser.parse_args()
    question = io_args.question
    
    if question == "3":        
        n = 5
        d = 4
        k = 3

        # Initialize entities randomly
        X = np.random.randn(n, d)
        w = np.random.randn(d)
        y = X@w + np.random.randn()

        # Declare constants and variables
        X_const = Constant(X)
        minus_y_const = Constant(-y)
        y_const = Constant(y)
        w_var = Variable(w)  # w_var.gradient will be automatically populated with backward()
        half_constant = Constant(np.array(0.5))

        # Compose functions
        f_var = 0.5 * nampy.sum((X_const @ w_var - y_const) ** 2)

        f = f_var.forward()  # forward() will evaluate the function and also maintain function object state for fast backward() compute
        f_var.backward()  # this populates w_var.gradient
        assert_almost_equal(f, 0.5 * np.sum((X@w - y)**2))
        assert_array_almost_equal(w_var.gradient, X.T@X@w - X.T@y)

        x = np.random.randn(d)
        x_var = Variable(x)
        x_const = Constant(x)
        f = x_var.forward()
        x_var.backward()
        assert_array_almost_equal(f, x)
        assert_array_almost_equal(x_var.gradient, np.eye(d))

        f_var = w_var.T @ x_const
        f = f_var.forward()
        f_var.backward()
        assert_array_almost_equal(f, w.T @ x)
        assert_array_almost_equal(w_var.gradient, x)

        f_var = X_const @ w_var
        f = f_var.forward()
        f_var.backward()
        assert_array_almost_equal(f, X @ w)
        assert_array_almost_equal(w_var.gradient, X.T)

        print("All tests passed.")

    elif question == "3.1":
        n = 5
        d = 4
        k = 3

        # Initialize entities randomly
        X = np.random.randn(n, d)
        w = np.random.randn(d)
        y = np.sign(X@w + np.random.randn())

        # Declare constants and variables
        X_const = Constant(X)
        y_const = Constant(y)
        ones_const = Constant(np.ones_like(y))
        w_var = Variable(w)  # w_var.gradient will be automatically populated with backward()

        """YOUR CODE HERE FOR Q3.1"""
        # TODO: compose differentiable functions to output a logistic loss based on X_const, y_const, and w_var
        f_var = None
        raise NotImplementedError()

        f = f_var.forward()  # forward() will evaluate the function and also maintain function object state for fast backward() compute
        f_var.backward()  # this populates w_var.gradient
        Xw = X @ w
        yXw = y * Xw  # element-wise multiply
        assert_almost_equal(f, np.sum(np.log(1 + np.exp(-yXw))))
        res = - y / (1. + np.exp(yXw))
        g = X.T @ res
        assert_array_almost_equal(w_var.gradient, g)

        print("All tests passed.")

    elif question == "3.2":
        n = 5
        d = 4
        k = 3

        # Initialize entities randomly
        X = np.random.randn(n, d)
        W = np.random.randn(k, d)
        w = np.random.randn(d)
        Y = X@W.T + np.random.randn()
        y = X@w + np.random.randn()
        y_binary = y > 0.5

        # Declare constants and variables
        X_const = Constant(X)
        Y_const = Constant(Y)
        y_const = Constant(y)
        minus_Y_const = Constant(-Y)
        W_var = Variable(W)  # w_var.gradient will be automatically populated with backward()
        w_var = Variable(w)
        y_binary_const = Constant(y_binary)

        # Sigmoid function
        # TODO: Implement sigmoid() in nampy.py
        f_var = 0.5 * nampy.sum((nampy.sigmoid(X_const @ w_var) - y_binary_const) ** 2)
        f = f_var.forward()
        f_var.backward()
        r = 1. / (1 + np.exp(-X@w)) - y_binary
        assert_array_almost_equal(f, 0.5 * np.sum(r**2))
        H = np.diag(1. / (1 + np.exp(-X@w)) * 1. / (1 + np.exp(X@w)))
        assert_array_almost_equal(w_var.gradient, X.T@H@r)
        print("Sigmoid tests passed.")

        # ReLU function
        # TODO: implment relu() in nampy.py
        f_var = 0.5 * nampy.sum((nampy.relu(X_const @ w_var) - y_const) ** 2)
        y_hat = X@w
        r = y_hat - y
        r[y_hat <= 0] = 0
        g = X.T@r
        f = f_var.forward()
        f_var.backward()
        assert_array_almost_equal(f, 0.5 * np.sum((np.maximum(0., y_hat) - y) ** 2))
        assert_array_almost_equal(w_var.gradient, g)
        print("ReLU tests passed.")

        # Matrix multiplication: if successfully implemented, multi-output linear model will work (ignore bias)
        f_var = 0.5 * nampy.sum((X_const @ W_var.T - Y_const)**2)
        f = f_var.forward()
        f_var.backward()  # Populates W_var.gradient
        assert_array_almost_equal(f, 0.5 * np.sum((X @ W.T - Y)**2))
        assert_array_almost_equal(W_var.gradient, np.transpose(X.T@(X@W.T - Y)))
        print("MatMul tests passed.")

        print("All tests passed.")

    elif question == "3.3":
        n = 100
        d = 10
        k = 3

        # Initialize entities randomly
        X = np.random.random([n, d])
        W = np.random.random([k, d])
        Y = X@W.T + np.random.randn()
        Y += np.ones(Y.shape)

        mapper = MapperLinear(d, k)
        optimizer = OptimizerGradientDescentLineSearch()
        fun_obj = FunObjLeastSquaresMultiOutput(mapper)  # with differentiable programming, objective function should have access to operations
        model = GenericSupervisedModel(mapper, optimizer, fun_obj)
        model.fit(X, Y)
        Y_hat = model.predict(X)
        err_train = np.mean((Y_hat - Y) ** 2)
        print("Training error: {:.3f}".format(err_train))

        assert err_train < 0.01
        print("All tests passed.")

    elif question == "3.4":
        n = 100
        d = 2
        k = 2
        q = 2

        # Initialize entities randomly
        X = np.random.randn(n, d)
        W = np.random.randn(k, d)
        V = np.random.randn(q, k)
        Y = ((X@W.T) > 0)
        
        hidden_dims = []
        output_dim = 32

        """YOUR CODE HERE FOR Q3.4"""
        raise NotImplementedError()
        # TODO: replace these variables with appropriate objects.
        mapper = None
        fun_obj = None

        child_optimizer = OptimizerGradientDescentHeavyBall(0.9)
        learning_rate_getter = LearningRateGetterConstant(1e-2)
        optimizer = OptimizerStochasticGradient(child_optimizer, learning_rate_getter, 10, max_evals=30)
        model = GenericSupervisedModel(mapper, optimizer, fun_obj)
        model.fit(X, Y)      

        Y_hat = model.predict(X)
        err_train = np.mean((Y_hat - Y) ** 2)

        print("Training error: {:.3f}".format(err_train))

        assert err_train < 0.08
        print("All tests passed.")


    else:
        print("Unknown question: %s" % question)    